--[[
Name: DrWho-1.0
Revision: $Rev: 32 $
Author(s): Rob Stiles (robs@codexsoftware.co.uk)
Website: http://www.wowace.com/
Documentation: http://www.wowace.com/index.php/DrWho-1.0
SVN: http://svn.wowace.com/wowace/trunk/DrWhoLib/DrWho-1.0
Description: Ace2 library for centralizing and caching the WoW /who lookup resources
Dependencies: AceLibrary, AceEvent-2.0, AceHook-2.1, AceOO-2.0
]]

local MAJOR_VERSION = "DrWho-1.0"
local MINOR_VERSION = 90000 + tonumber(("$Revision: 32 $"):match("(%d+)"))

-- Make sure we have our dependencies
if not AceLibrary then error(MAJOR_VERSION .. " requires AceLibrary.") end
if not AceLibrary:HasInstance('AceEvent-2.0') then error(MAJOR_VERSION .. ' requires AceEvent-2.0.') end
if not AceLibrary:HasInstance('AceHook-2.1') then error(MAJOR_VERSION .. ' requires AceHook-2.1') end
if not AceLibrary:HasInstance('AceOO-2.0') then error(MAJOR_VERSION .. ' requires AceOO-2.0') end
-- Make sure we're the newest version of this library or exit
if not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then return end

local AceOO = AceLibrary:GetInstance('AceOO-2.0')

local DrWho = {}

DrWho.debug = false

function DrWho:Debug(msg)
	if not self.debug then
		return
	end
	DEFAULT_CHAT_FRAME:AddMessage('DrWho Debug: ' .. msg)		
end


function DrWho:GetCachedWhoData(name)
	-- returns nil if there is none
	return self.whoCache[name]
end

-- maxAge is in seconds.  -1 = return cache info only.
function DrWho:GetWhoData(name, maxAge)
	if not name then
		return
	end
	
	-- We don't want to who lookup cross server players
	if string.find(name, "-") or string.find(name, "*") then
		self:Debug("Not going to do a who look up on a cross server player.")
		return
	end
	
	-- Let's not bother if the name contains a number either
	if string.match(name, "%d+") then
		self:Debug("Not going to do a who look up on a number.")
		return
	end

	if not maxAge then
		maxAge = 0
	end
	
	local now = time()
	local data = self:GetCachedWhoData(name)
	if data then
		if (now - data.lastUpdate < maxAge) or (maxAge == -1) or (maxAge == 0) then
			-- return a cached version
			return data
		else
			self:Debug("Data for " .. name .. " is out of date.")
		end
	end
	
	if maxAge ~= -1 then
		self:AddToWhoQ(name)
	end	
	return nil
end

function DrWho:AddToWhoQ(name)	
	if not self.whoQ[name] then
		local data = {}
		data.timeAdded  = time()
		self.whoQ[name] = data
		-- Trigger it now so we don't wait if we don't have to
		self:WhoMaintenance()
	end		
end

function DrWho:WhoMaintenance()	
	-- Time out look ups from the Queue that have been in there too long
	local timeOut = 120
	local now = time()
	local count = 0
	for name, data in pairs(self.whoQ) do
		self:Debug("Checking " .. name .. " for timeout.")	
		if now - self.whoQ[name].timeAdded > timeOut then
			-- This one has been on retry long enough
			self:Debug("Timing out " .. name)
			self.whoQ[name] = nil
		else
			count = count +1
		end
	end
			
	if self.lastQIndex >= count then
		self.lastQIndex = 0
	end
	local index = 1	
	for name, data in pairs(self.whoQ) do		
		if index > self.lastQIndex then
			self.lastQIndex = index
			local string = "n-\"" .. name .. "\""
			self:Debug("Sending who lookup: " .. string)			
			self:SendWho(string, 1)
			break
		end	
		index = index +1		
	end		
end

function DrWho:WHO_LIST_UPDATE()
	local totalCount, numWhos = GetNumWhoResults()
	for i=1,numWhos do
		local name, guild, level, race, class, zone = GetWhoInfo(i)
		self:Debug("Adding " .. name .. " to internal cache.")	
		-- update our cache
		local data = {}
		data.guild = guild
		data.level = level
		data.race = race
		data.class = class
		data.zone = zone
		data.lastUpdate = time()
		self.whoCache[name] = data		
		self.whoQ[name] = nil
	end
	
	local qEmpty = true
	for name, data in pairs(self.whoQ) do
		qEmpty = false
	end
	
	if qEmpty then
		if self:GetOldWhoToUI() ~= self:GetWhoToUI() then
			self:Debug("Restoring old WhoToUI")
			self:SetWhoToUI(self:GetOldWhoToUI())
		end
	end			
end

function DrWho:FriendsFrame_OnEvent()
	local callHook = true
	if event == "WHO_LIST_UPDATE" then
		local lib = AceLibrary("DrWho-1.0")
		lib:Debug("FriendsFrame_OnEvent() called.")
		local totalCount, numWhos = GetNumWhoResults()		
		
		
		for whoQName, data in pairs(lib.whoQ) do
			-- We have something in our queue
			if numWhos == 0 then
				-- and we had an empty who list back
				callHook = false
			end			
			break
		end
						
		for i=1,numWhos do
			local name, guild, level, race, class, zone = GetWhoInfo(i)
			-- Check to see if it's one that we've requested
			for whoQName, data in pairs(lib.whoQ) do
				lib:Debug("String matching " .. string.lower(name) .. " with " .. string.lower(whoQName))
				if string.match(string.lower(name), string.lower(whoQName)) then
					-- This who response almost certainly belongs to us					
					callHook = false
					lib.whoQ[whoQName] = nil
					break
				end
			end						
		end	
	end	
	if callHook then
		self.hooks["FriendsFrame_OnEvent"](event)
	end	
end

-- Something that should be in the WoW API imho
function DrWho:GetWhoToUI()
	return self.whoToUI
end

function DrWho:GetOldWhoToUI()
	return self.oldWhoToUI
end

function DrWho:SetWhoToUI(newVal, internal)
	local t = "true"
	if not internal then t = "false" end
	self:Debug("SetWhoToUI(" .. newVal .. ", " .. t ..")")
	if not internal then
		self.oldWhoToUI = newVal
	end
	self.whoToUI = newVal
	self.hooks.SetWhoToUI(newVal)
end

function DrWho:SendWho(msg, toUI)
	local t
	if not toUI then 
		t = "false"
	else
		t = "true"
	end
	self:Debug("SendWho(" .. msg .. ", " .. t .. ")")
	if toUI ~= nil then
		self:SetWhoToUI(toUI, true)
	end	
	self.hooks.SendWho(msg)
end


local function activate(self, oldLib, oldDeactivate)
    if oldLib then -- if upgrading
        self.GetWhoData = oldLib.GetWhoData
    end
            
    -- make sure all the basic values are available
	if not self.whoCache then
		self.whoCache = {}
	end
	if not self.whoQ then
		self.whoQ = {}
	end
    if not self.lastQIndex then
		self.lastQIndex = 0
    end
    
    if not self.oldWhoToUI then
		self.oldWhoToUI = 0
    end

    if not self.whoToUI then
		self.whoToUI = 0
    end
    
    --self.debug = true
        
    if oldDeactivate then -- clean up the old library
        oldDeactivate(oldLib)
    end
end

local function deactivate(oldLib)
	oldLib:CancelAllScheduledEvents()
	oldLib:UnregisterAllEvents()
	oldLib:UnhookAll()
end

local function external(self, major, instance)
	if(major == 'AceEvent-2.0')then
		instance:embed(self)
		self:CancelAllScheduledEvents()		
		self:RegisterEvent('WHO_LIST_UPDATE')
		self:ScheduleRepeatingEvent(self.WhoMaintenance, 6, self)
	elseif(major == 'AceHook-2.1')then
		instance:embed(self)
		self:Hook('SendWho', true)		
		self:Hook('FriendsFrame_OnEvent', true)
		self:Hook('SetWhoToUI', true)		
	end
end

AceLibrary:Register(DrWho, MAJOR_VERSION, MINOR_VERSION, activate, deactivate, external)

